import asyncio
from typing import Any, cast

from arcade_tdk import ToolContext
from arcade_tdk.errors import ToolExecutionError
from slack_sdk.errors import SlackApiError
from slack_sdk.web.async_client import AsyncWebClient

from arcade_slack.constants import MAX_CONCURRENT_REQUESTS, MAX_PAGINATION_TIMEOUT_SECONDS
from arcade_slack.models import (
    FindMultipleUsersByIdSentinel,
    FindMultipleUsersByUsernameSentinel,
    GetUserByEmailCaller,
)
from arcade_slack.utils import (
    async_paginate,
    build_multiple_users_retrieval_response,
    cast_user_dict,
    gather_with_concurrency_limit,
    is_user_a_bot,
    is_valid_email,
    short_user_info,
)


async def get_users_by_id_username_or_email(
    context: ToolContext,
    user_ids: str | list[str] | None = None,
    usernames: str | list[str] | None = None,
    emails: str | list[str] | None = None,
    semaphore: asyncio.Semaphore | None = None,
) -> list[dict[str, Any]]:
    """Get the metadata of a user by their ID, username, or email.

    Provide any combination of user_ids, usernames, and/or emails. Always prefer providing user_ids
    and/or emails, when available, since the performance is better.
    """
    if isinstance(user_ids, str):
        user_ids = [user_ids]
    if isinstance(usernames, str):
        usernames = [usernames]
    if isinstance(emails, str):
        emails = [emails]

    if not any([user_ids, usernames, emails]):
        raise ToolExecutionError("At least one of user_ids, usernames, or emails must be provided")

    if not semaphore:
        semaphore = asyncio.Semaphore(MAX_CONCURRENT_REQUESTS)

    user_retrieval_calls = []

    auth_token = context.get_auth_token_or_empty()

    if user_ids:
        user_retrieval_calls.append(get_users_by_id(auth_token, user_ids, semaphore))

    if usernames:
        user_retrieval_calls.append(get_users_by_username(auth_token, usernames, semaphore))

    if emails:
        user_retrieval_calls.append(get_users_by_email(auth_token, emails, semaphore))

    responses = await asyncio.gather(*user_retrieval_calls)

    return await build_multiple_users_retrieval_response(context, responses)


async def get_users_by_id(
    auth_token: str,
    user_ids: list[str],
    semaphore: asyncio.Semaphore | None = None,
) -> dict[str, list]:
    user_ids = list(set(user_ids))

    if len(user_ids) == 1:
        user = await get_single_user_by_id(auth_token, user_ids[0])
        if not user:
            return {"users": [], "not_found": user_ids}
        else:
            return {"users": [user], "not_found": []}

    if not semaphore:
        semaphore = asyncio.Semaphore(MAX_CONCURRENT_REQUESTS)

    async with semaphore:
        slack_client = AsyncWebClient(token=auth_token)
        response, _ = await async_paginate(
            func=slack_client.users_list,
            response_key="members",
            sentinel=FindMultipleUsersByIdSentinel(user_ids=user_ids),
        )

    user_ids_pending = set(user_ids)
    users = []

    for user in response:
        user_dict = cast(dict, user)
        if user_dict["id"] in user_ids_pending:
            users.append(cast_user_dict(user_dict))
            user_ids_pending.remove(user_dict["id"])

    return {"users": users, "not_found": list(user_ids_pending)}


async def get_single_user_by_id(auth_token: str, user_id: str) -> dict[str, Any] | None:
    slack_client = AsyncWebClient(token=auth_token)
    try:
        response = await slack_client.users_info(user=user_id)
        if not response.get("ok"):
            return None
        return cast_user_dict(response["user"])
    except SlackApiError as e:
        if "not_found" in e.response.get("error", ""):
            return None
        else:
            message = f"There was an error getting the user with ID {user_id}."
            slack_error_message = e.response.get("error", "Unknown Slack API error")
            raise ToolExecutionError(
                message=message,
                developer_message=f"{message}: {slack_error_message}",
            ) from e


async def get_users_by_username(
    auth_token: str,
    usernames: list[str],
    semaphore: asyncio.Semaphore | None = None,
) -> dict[str, list[dict]]:
    usernames = list(set(usernames))

    if not semaphore:
        semaphore = asyncio.Semaphore(MAX_CONCURRENT_REQUESTS)

    slack_client = AsyncWebClient(token=auth_token)

    async with semaphore:
        users, _ = await async_paginate(
            func=slack_client.users_list,
            response_key="members",
            max_pagination_timeout_seconds=MAX_PAGINATION_TIMEOUT_SECONDS,
            sentinel=FindMultipleUsersByUsernameSentinel(usernames=usernames),
        )

    users_found = []
    usernames_lower = {username.casefold() for username in usernames}
    usernames_pending = set(usernames)
    available_users = []

    for user in users:
        if is_user_a_bot(user):
            continue

        available_users.append(short_user_info(user))

        if not isinstance(user.get("name"), str):
            continue

        username_lower = user["name"].casefold()

        if username_lower in usernames_lower:
            users_found.append(cast_user_dict(user))
            # Username/handle is unique in Slack, we can ignore it after finding a match
            for pending_username in usernames_pending:
                if pending_username.casefold() == username_lower:
                    usernames_pending.remove(pending_username)
                    break

    response: dict[str, Any] = {"users": users_found}

    if usernames_pending:
        response["not_found"] = list(usernames_pending)
        response["available_users"] = available_users

    return response


async def get_users_by_email(
    auth_token: str,
    emails: list[str],
    semaphore: asyncio.Semaphore | None = None,
) -> dict[str, list[dict]]:
    emails = list(set(emails))

    for email in emails:
        if not is_valid_email(email):
            raise ToolExecutionError(f"Invalid email address: {email}")

    if not semaphore:
        semaphore = asyncio.Semaphore(MAX_CONCURRENT_REQUESTS)

    slack_client = AsyncWebClient(token=auth_token)
    callers = [GetUserByEmailCaller(slack_client.users_lookupByEmail, email) for email in emails]

    results = await gather_with_concurrency_limit(
        coroutine_callers=callers,
        semaphore=semaphore,
    )

    users = []
    emails_not_found = []

    for result in results:
        if result["user"]:
            users.append(cast_user_dict(result["user"]))
        else:
            emails_not_found.append(result["email"])

    response: dict[str, Any] = {"users": users}

    if emails_not_found:
        response["not_found"] = emails_not_found

    return response



from datetime import timedelta

from arcade_evals import (
    BinaryCritic,
    DatetimeCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_slack
from arcade_slack.critics import RelativeTimeBinaryCritic
from arcade_slack.tools.chat import get_messages

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.8,
    warn_threshold=0.9,
)


catalog = ToolCatalog()
# Register the Slack tools
catalog.add_module(arcade_slack)


@tool_eval()
def get_messages_in_direct_message_eval_suite() -> EvalSuite:
    """Create an evaluation suite for tools getting messages in direct messages."""
    suite = EvalSuite(
        name="Slack Chat Tools Evaluation",
        system_message="You are an AI assistant that can interact with Slack to send messages and get information from conversations, users, etc.",
        catalog=catalog,
        rubric=rubric,
    )

    no_arguments_user_messages_by_username = [
        "what are the latest messages I exchanged with janedoe",
        "show my messages with janedoe on Slack",
        "get the messages I exchanged with janedoe",
        "get the message history with janedoe",
    ]

    for i, user_message in enumerate(no_arguments_user_messages_by_username):
        suite.add_case(
            name=f"{user_message} [{i}]",
            user_message=user_message,
            expected_tool_calls=[
                ExpectedToolCall(
                    func=get_messages,
                    args={
                        "conversation_id": None,
                        "channel_name": None,
                        "user_ids": None,
                        "usernames": ["janedoe"],
                        "emails": None,
                    },
                ),
            ],
            critics=[
                BinaryCritic(critic_field="conversation_id", weight=0.1),
                BinaryCritic(critic_field="channel_name", weight=0.1),
                BinaryCritic(critic_field="user_ids", weight=0.1),
                BinaryCritic(critic_field="usernames", weight=0.6),
                BinaryCritic(critic_field="emails", weight=0.1),
            ],
        )

    no_arguments_user_messages_by_email = [
        "what are the latest messages I exchanged with jane.doe@acme.com",
        "show my messages with jane.doe@acme.com on Slack",
    ]

    for i, user_message in enumerate(no_arguments_user_messages_by_email):
        suite.add_case(
            name=f"{user_message} [{i}]",
            user_message=user_message,
            expected_tool_calls=[
                ExpectedToolCall(
                    func=get_messages,
                    args={
                        "conversation_id": None,
                        "channel_name": None,
                        "user_ids": None,
                        "usernames": None,
                        "emails": ["jane.doe@acme.com"],
                    },
                ),
            ],
            critics=[
                BinaryCritic(critic_field="conversation_id", weight=0.1),
                BinaryCritic(critic_field="channel_name", weight=0.1),
                BinaryCritic(critic_field="user_ids", weight=0.1),
                BinaryCritic(critic_field="usernames", weight=0.1),
                BinaryCritic(critic_field="emails", weight=0.6),
            ],
        )

    suite.add_case(
        name="get messages in direct conversation by username (on a specific date)",
        user_message="get the messages I exchanged with janedoe on 2025-01-31",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages,
                args={
                    "conversation_id": None,
                    "channel_name": None,
                    "user_ids": None,
                    "usernames": ["janedoe"],
                    "emails": None,
                    "oldest_datetime": "2025-01-31 00:00:00",
                    "latest_datetime": "2025-01-31 23:59:59",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=0.1),
            BinaryCritic(critic_field="channel_name", weight=0.1),
            BinaryCritic(critic_field="user_ids", weight=0.1),
            BinaryCritic(critic_field="usernames", weight=0.2),
            BinaryCritic(critic_field="emails", weight=0.1),
            DatetimeCritic(
                critic_field="oldest_datetime", weight=0.2, max_difference=timedelta(minutes=2)
            ),
            DatetimeCritic(
                critic_field="latest_datetime", weight=0.2, max_difference=timedelta(minutes=2)
            ),
        ],
    )

    suite.add_case(
        name="get messages in direct conversation by email (on a specific date)",
        user_message="get the messages I exchanged with jane.doe@acme.com on 2025-01-31",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages,
                args={
                    "conversation_id": None,
                    "channel_name": None,
                    "user_ids": None,
                    "usernames": None,
                    "emails": ["jane.doe@acme.com"],
                    "oldest_datetime": "2025-01-31 00:00:00",
                    "latest_datetime": "2025-01-31 23:59:59",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=0.1),
            BinaryCritic(critic_field="channel_name", weight=0.1),
            BinaryCritic(critic_field="user_ids", weight=0.1),
            BinaryCritic(critic_field="usernames", weight=0.1),
            BinaryCritic(critic_field="emails", weight=0.2),
            DatetimeCritic(
                critic_field="oldest_datetime", weight=0.2, max_difference=timedelta(minutes=2)
            ),
            DatetimeCritic(
                critic_field="latest_datetime", weight=0.2, max_difference=timedelta(minutes=2)
            ),
        ],
    )

    suite.add_case(
        name="Get conversation history oldest relative by username (2 days ago)",
        user_message="Get the messages I exchanged with janedoe starting 2 days ago",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_messages,
                args={
                    "conversation_id": None,
                    "channel_name": None,
                    "user_ids": None,
                    "usernames": ["janedoe"],
                    "emails": None,
                    "oldest_relative": "02:00:00",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=0.1),
            BinaryCritic(critic_field="channel_name", weight=0.1),
            BinaryCritic(critic_field="user_ids", weight=0.1),
            BinaryCritic(critic_field="usernames", weight=0.3),
            BinaryCritic(critic_field="emails", weight=0.1),
            RelativeTimeBinaryCritic(critic_field="oldest_relative", weight=0.3),
        ],
    )

    return suite



import asyncio
import copy
import json
from unittest.mock import AsyncMock, call, patch

import pytest
from arcade_tdk.errors import RetryableToolError
from slack_sdk.errors import SlackApiError
from slack_sdk.web.async_client import AsyncWebClient

from arcade_slack.exceptions import PaginationTimeoutError
from arcade_slack.models import (
    ConcurrencySafeCoroutineCaller,
    FindMultipleUsersByUsernameSentinel,
    FindUserByUsernameSentinel,
)
from arcade_slack.utils import (
    async_paginate,
    build_multiple_users_retrieval_response,
    filter_conversations_by_user_ids,
    gather_with_concurrency_limit,
    is_valid_email,
    populate_users_in_messages,
)


@pytest.mark.asyncio
async def test_async_paginate():
    mock_slack_client = AsyncMock()
    mock_slack_client.conversations_list.return_value = {
        "ok": True,
        "channels": [{"id": "123"}],
        "response_metadata": {"next_cursor": None},
    }

    results, next_cursor = await async_paginate(
        func=mock_slack_client.conversations_list,
        response_key="channels",
    )

    assert results == [{"id": "123"}]
    assert next_cursor is None


@pytest.mark.asyncio
async def test_async_paginate_with_find_user_sentinel():
    mock_slack_client = AsyncMock()
    mock_slack_client.users_list.side_effect = [
        {
            "ok": True,
            "members": [
                {"id": "123", "name": "Jack"},
                {"id": "456", "name": "John"},
            ],
            "response_metadata": {"next_cursor": "cursor1"},
        },
        {
            "ok": True,
            "members": [{"id": "789", "name": "Jenifer"}],
            "response_metadata": {"next_cursor": "cursor2"},
        },
        {
            "ok": True,
            "members": [{"id": "007", "name": "James"}],
            "response_metadata": {"next_cursor": None},
        },
    ]

    results, next_cursor = await async_paginate(
        func=mock_slack_client.users_list,
        response_key="members",
        sentinel=FindUserByUsernameSentinel(username="jenifer"),
    )

    assert results == [
        {"id": "123", "name": "Jack"},
        {"id": "456", "name": "John"},
        {"id": "789", "name": "Jenifer"},
    ]
    assert next_cursor == "cursor2"


@pytest.mark.asyncio
async def test_async_paginate_with_find_user_sentinel_not_found():
    mock_slack_client = AsyncMock()
    mock_slack_client.users_list.side_effect = [
        {
            "ok": True,
            "members": [
                {"id": "123", "name": "Jack"},
                {"id": "456", "name": "John"},
            ],
            "response_metadata": {"next_cursor": "cursor1"},
        },
        {
            "ok": True,
            "members": [{"id": "789", "name": "Jenifer"}],
            "response_metadata": {"next_cursor": "cursor2"},
        },
        {
            "ok": True,
            "members": [{"id": "007", "name": "James"}],
            "response_metadata": {"next_cursor": None},
        },
    ]

    results, next_cursor = await async_paginate(
        func=mock_slack_client.users_list,
        response_key="members",
        sentinel=FindUserByUsernameSentinel(username="Do not find me"),
    )

    assert results == [
        {"id": "123", "name": "Jack"},
        {"id": "456", "name": "John"},
        {"id": "789", "name": "Jenifer"},
        {"id": "007", "name": "James"},
    ]
    assert next_cursor is None


@pytest.mark.asyncio
async def test_async_paginate_with_find_multiple_users_sentinel():
    mock_slack_client = AsyncMock()
    mock_slack_client.users_list.side_effect = [
        {
            "ok": True,
            "members": [
                {"id": "123", "name": "Jack"},
                {"id": "456", "name": "John"},
            ],
            "response_metadata": {"next_cursor": "cursor1"},
        },
        {
            "ok": True,
            "members": [
                {"id": "789", "name": "Jenifer"},
                {"id": "101", "name": "Janis"},
            ],
            "response_metadata": {"next_cursor": "cursor2"},
        },
        {
            "ok": True,
            "members": [{"id": "007", "name": "James"}],
            "response_metadata": {"next_cursor": None},
        },
    ]

    results, next_cursor = await async_paginate(
        func=mock_slack_client.users_list,
        response_key="members",
        sentinel=FindMultipleUsersByUsernameSentinel(usernames=["jenifer", "jack"]),
    )

    assert results == [
        {"id": "123", "name": "Jack"},
        {"id": "456", "name": "John"},
        {"id": "789", "name": "Jenifer"},
        {"id": "101", "name": "Janis"},
    ]
    assert next_cursor == "cursor2"


@pytest.mark.asyncio
async def test_async_paginate_with_find_multiple_users_sentinel_not_found():
    mock_slack_client = AsyncMock()
    mock_slack_client.users_list.side_effect = [
        {
            "ok": True,
            "members": [
                {"id": "123", "name": "Jack"},
                {"id": "456", "name": "John"},
            ],
            "response_metadata": {"next_cursor": "cursor1"},
        },
        {
            "ok": True,
            "members": [
                {"id": "789", "name": "Jenifer"},
                {"id": "101", "name": "Janis"},
            ],
            "response_metadata": {"next_cursor": "cursor2"},
        },
        {
            "ok": True,
            "members": [{"id": "007", "name": "James"}],
            "response_metadata": {"next_cursor": None},
        },
    ]

    results, next_cursor = await async_paginate(
        func=mock_slack_client.users_list,
        response_key="members",
        sentinel=FindMultipleUsersByUsernameSentinel(
            usernames=["jenifer", "jack", "do not find me"]
        ),
    )

    assert results == [
        {"id": "123", "name": "Jack"},
        {"id": "456", "name": "John"},
        {"id": "789", "name": "Jenifer"},
        {"id": "101", "name": "Janis"},
        {"id": "007", "name": "James"},
    ]
    assert next_cursor is None


@pytest.mark.asyncio
async def test_async_paginate_with_response_error():
    mock_slack_client = AsyncMock()
    mock_slack_client.conversations_list.side_effect = SlackApiError(
        message="slack_error",
        response={"ok": False, "error": "slack_error"},
    )

    with pytest.raises(SlackApiError) as e:
        await async_paginate(
            func=mock_slack_client.conversations_list,
            response_key="channels",
        )
        assert str(e.value) == "slack_error"


@pytest.mark.asyncio
async def test_async_paginate_with_custom_pagination_args():
    mock_slack_client = AsyncMock()
    mock_slack_client.conversations_list.return_value = {
        "ok": True,
        "channels": [{"id": "123"}],
        "response_metadata": {"next_cursor": "456"},
    }

    results, next_cursor = await async_paginate(
        func=mock_slack_client.conversations_list,
        response_key="channels",
        limit=1,
        next_cursor="123",
        hello="world",
    )

    assert results == [{"id": "123"}]
    assert next_cursor == "456"

    mock_slack_client.conversations_list.assert_called_once_with(
        hello="world",
        limit=1,
        cursor="123",
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "test_limit, last_next_cursor, last_expected_limit",
    [(5, "cursor3", 1), (None, None, 2)],
)
async def test_async_paginate_large_limit(test_limit, last_next_cursor, last_expected_limit):
    mock_slack_client = AsyncMock(spec=AsyncWebClient)
    mock_slack_client.conversations_list.side_effect = [
        {
            "ok": True,
            "channels": [{"id": "channel1"}, {"id": "channel2"}],
            "response_metadata": {"next_cursor": "cursor1"},
        },
        {
            "ok": True,
            "channels": [{"id": "channel3"}, {"id": "channel4"}],
            "response_metadata": {"next_cursor": "cursor2"},
        },
        {
            "ok": True,
            "channels": [{"id": "channel5"}],
            "response_metadata": {"next_cursor": last_next_cursor},
        },
    ]

    with patch("arcade_slack.utils.MAX_PAGINATION_SIZE_LIMIT", 2):
        results, next_cursor = await async_paginate(
            func=mock_slack_client.conversations_list,
            response_key="channels",
            limit=test_limit,
            hello="world",
        )

    assert results == [
        {"id": "channel1"},
        {"id": "channel2"},
        {"id": "channel3"},
        {"id": "channel4"},
        {"id": "channel5"},
    ]
    assert next_cursor == last_next_cursor
    assert mock_slack_client.conversations_list.call_count == 3
    mock_slack_client.conversations_list.assert_has_calls([
        call(hello="world", limit=2, cursor=None),
        call(hello="world", limit=2, cursor="cursor1"),
        call(hello="world", limit=last_expected_limit, cursor="cursor2"),
    ])


@pytest.mark.asyncio
async def test_async_paginate_large_limit_with_response_error():
    mock_slack_client = AsyncMock()
    mock_slack_client.conversations_list.side_effect = [
        {
            "ok": True,
            "channels": [{"id": "channel1"}, {"id": "channel2"}],
            "response_metadata": {"next_cursor": "cursor1"},
        },
        SlackApiError(message="slack_error", response={"ok": False, "error": "slack_error"}),
        {
            "ok": True,
            "channels": [{"id": "channel5"}],
            "response_metadata": {"next_cursor": "cursor3"},
        },
    ]

    with (
        patch("arcade_slack.utils.MAX_PAGINATION_SIZE_LIMIT", 2),
        pytest.raises(SlackApiError) as e,
    ):
        await async_paginate(
            func=mock_slack_client.conversations_list,
            response_key="channels",
            limit=5,
            hello="world",
        )
        assert str(e.value) == "slack_error"

    assert mock_slack_client.conversations_list.call_count == 2
    mock_slack_client.conversations_list.assert_has_calls([
        call(hello="world", limit=2, cursor=None),
        call(hello="world", limit=2, cursor="cursor1"),
    ])


@pytest.mark.asyncio
async def test_async_paginate_with_timeout():
    # Mock Slack client
    mock_slack_client = AsyncMock()

    # Simulate a network delay by making the mock function sleep
    async def mock_conversations_list(*args, **kwargs):
        await asyncio.sleep(1)  # Sleep for 1 second to simulate delay
        return {
            "ok": True,
            "channels": [{"id": "123"}],
            "response_metadata": {"next_cursor": None},
        }

    mock_slack_client.conversations_list.side_effect = mock_conversations_list

    # Set a low timeout to trigger the timeout error quickly during the test
    max_pagination_timeout_seconds = 0.1  # 100 milliseconds

    with pytest.raises(PaginationTimeoutError) as exc_info:
        await async_paginate(
            func=mock_slack_client.conversations_list,
            response_key="channels",
            max_pagination_timeout_seconds=max_pagination_timeout_seconds,
        )

    assert (
        str(exc_info.value)
        == f"The pagination process timed out after {max_pagination_timeout_seconds} seconds."
    )


def test_filter_conversations_by_user_ids():
    conversations = [
        {"id": "123", "members": [{"id": "user1"}, {"id": "user2"}, {"id": "user3"}]},
        {"id": "456", "members": [{"id": "user2"}, {"id": "user3"}]},
    ]
    response = filter_conversations_by_user_ids(
        conversations=conversations,
        user_ids=["user1", "user2"],
        exact_match=False,
    )
    assert response == [
        {"id": "123", "members": [{"id": "user1"}, {"id": "user2"}, {"id": "user3"}]},
    ]


def test_filter_conversations_by_user_ids_empty_response():
    conversations = [
        {"id": "123", "members": [{"id": "user1"}, {"id": "user3"}, {"id": "user4"}]},
        {"id": "456", "members": [{"id": "user2"}, {"id": "user3"}, {"id": "user4"}]},
    ]
    response = filter_conversations_by_user_ids(
        conversations=conversations,
        user_ids=["user1", "user2"],
        exact_match=False,
    )
    assert response == []


def test_filter_conversations_by_user_ids_multiple_matches():
    conversations = [
        {"id": "123", "members": [{"id": "user1"}, {"id": "user2"}, {"id": "user3"}]},
        {"id": "456", "members": [{"id": "user2"}, {"id": "user3"}]},
        {
            "id": "789",
            "members": [{"id": "user4"}, {"id": "user1"}, {"id": "user2"}, {"id": "user3"}],
        },
    ]
    response = filter_conversations_by_user_ids(
        conversations=conversations,
        user_ids=["user1", "user2"],
        exact_match=False,
    )
    assert response == [
        {"id": "123", "members": [{"id": "user1"}, {"id": "user2"}, {"id": "user3"}]},
        {
            "id": "789",
            "members": [{"id": "user4"}, {"id": "user1"}, {"id": "user2"}, {"id": "user3"}],
        },
    ]


def test_filter_conversations_by_user_ids_exact_match():
    conversations = [
        {"id": "123", "members": [{"id": "user1"}, {"id": "user2"}]},
        {"id": "456", "members": [{"id": "user2"}, {"id": "user3"}]},
    ]
    response = filter_conversations_by_user_ids(
        conversations=conversations,
        user_ids=["user1", "user2"],
        exact_match=True,
    )
    assert response == [{"id": "123", "members": [{"id": "user1"}, {"id": "user2"}]}]


def test_filter_conversations_by_user_ids_exact_match_empty_response():
    conversations = [
        {"id": "123", "members": [{"id": "user1"}, {"id": "user2"}, {"id": "user3"}]},
        {"id": "456", "members": [{"id": "user2"}, {"id": "user3"}]},
    ]
    response = filter_conversations_by_user_ids(
        conversations=conversations,
        user_ids=["user1", "user2"],
        exact_match=True,
    )
    assert response == []


@pytest.mark.parametrize(
    "users_by_email, users_by_username, expected_response",
    [
        (
            {"users": [{"id": "U1", "name": "user1"}]},
            {"users": [{"id": "U2", "name": "user2"}]},
            [{"id": "U1", "name": "user1"}, {"id": "U2", "name": "user2"}],
        ),
        (
            {"users": [{"id": "U1", "name": "user1"}]},
            {"users": []},
            [{"id": "U1", "name": "user1"}],
        ),
        (
            {"users": []},
            {"users": [{"id": "U2", "name": "user2"}]},
            [{"id": "U2", "name": "user2"}],
        ),
        (
            {"users": []},
            {"users": []},
            [],
        ),
    ],
)
@pytest.mark.asyncio
async def test_build_multiple_users_retrieval_response_success(
    users_by_email,
    users_by_username,
    expected_response,
    mock_context,
):
    response = await build_multiple_users_retrieval_response(
        context=mock_context,
        users_responses=[users_by_email, users_by_username],
    )
    assert response == expected_response


@pytest.mark.parametrize(
    "users_by_email, users_by_username",
    [
        # Both emails and usernames not found
        (
            {
                "users": [{"id": "U1", "name": "user1"}],
                "not_found": ["email_not_found"],
            },
            {
                "users": [{"id": "U2", "name": "user2"}],
                "not_found": ["username_not_found"],
                "available_users": [{"id": "U3", "name": "user3"}],
            },
        ),
        # Email not found, usernames found
        (
            {
                "users": [{"id": "U1", "name": "user1"}],
                "not_found": ["email_not_found"],
            },
            {
                "users": [{"id": "U2", "name": "user2"}],
                "not_found": [],
            },
        ),
        # Email found, username not found
        (
            {
                "users": [{"id": "U1", "name": "user1"}],
                "not_found": [],
            },
            {
                "users": [{"id": "U2", "name": "user2"}],
                "not_found": ["username_not_found"],
                "available_users": [{"id": "U3", "name": "user3"}],
            },
        ),
    ],
)
@pytest.mark.asyncio
async def test_build_multiple_users_retrieval_response_not_found(
    users_by_email,
    users_by_username,
    mock_context,
):
    with pytest.raises(RetryableToolError) as error:
        await build_multiple_users_retrieval_response(
            context=mock_context,
            users_responses=[users_by_email, users_by_username],
        )

    emails_not_found = users_by_email.get("not_found", [])
    usernames_not_found = users_by_username.get("not_found", [])
    available_users = users_by_username.get("available_users", [])

    for email in emails_not_found:
        assert email in error.value.message
    for username in usernames_not_found:
        assert username in error.value.message
    for user in available_users:
        assert json.dumps(user) in error.value.additional_prompt_content


def test_is_valid_email():
    assert is_valid_email("test@example.com")
    assert is_valid_email("test+123@example.com")
    assert is_valid_email("test-123@example.com")
    assert is_valid_email("test_123@example.com")
    assert is_valid_email("test.123@example.com")
    assert is_valid_email("test123@example.com")
    assert is_valid_email("test@example.co")
    assert is_valid_email("test@example.com.co")
    assert not is_valid_email("test123@example")
    assert not is_valid_email("test@example")
    assert not is_valid_email("test@example.")
    assert not is_valid_email("test@.com")
    assert not is_valid_email("test@example.c")
    assert not is_valid_email("test@example.com.")
    assert not is_valid_email("test@example.com.c")


@pytest.mark.asyncio
async def test_gather_with_concurrency_limit():
    mock_func1 = AsyncMock()
    mock_func2 = AsyncMock()

    caller1 = ConcurrencySafeCoroutineCaller(mock_func1, "arg1", "arg2", kwarg1="kwarg1")
    caller2 = ConcurrencySafeCoroutineCaller(mock_func2, "arg1", "arg2", kwarg1="kwarg1")

    mock_semaphore = AsyncMock(spec=asyncio.Semaphore)

    response = await gather_with_concurrency_limit(
        coroutine_callers=[caller1, caller2],
        semaphore=mock_semaphore,
    )

    response = tuple(response)

    assert len(response) == 2
    assert response[0] == mock_func1.return_value
    assert response[1] == mock_func2.return_value

    mock_func1.assert_awaited_once_with("arg1", "arg2", kwarg1="kwarg1")
    mock_func2.assert_awaited_once_with("arg1", "arg2", kwarg1="kwarg1")

    assert mock_semaphore.__aenter__.await_count == 2
    assert mock_semaphore.__aexit__.await_count == 2


@pytest.mark.asyncio
async def test_populate_users_in_messages(
    mock_context,
    mock_user_retrieval_slack_client,
    dummy_message_factory,
    dummy_reaction_factory,
    dummy_user_factory,
):
    user1 = dummy_user_factory(id_="U1", name="user1")
    user2 = dummy_user_factory(id_="U2", name="user2")
    user3 = dummy_user_factory(id_="U3", name="user3")
    user4 = dummy_user_factory(id_="U4", name="user4")
    user5 = dummy_user_factory(id_="U5", name="user5")

    user1_short = {"id": user1["id"], "name": user1["name"]}
    user2_short = {"id": user2["id"], "name": user2["name"]}
    user3_short = {"id": user3["id"], "name": user3["name"]}
    user4_short = {"id": user4["id"], "name": user4["name"]}

    user2_mention = f"<@{user2['name']} (id:{user2['id']})>"
    user5_mention = f"<@{user5['name']} (id:{user5['id']})>"

    reactions = [
        dummy_reaction_factory(name="thumbsup", user_ids=[user1["id"], user2["id"]]),
        dummy_reaction_factory(name="partyparrot", user_ids=[user3["id"], user4["id"]]),
    ]

    messages = [
        dummy_message_factory(
            user_id=user1["id"],
            text=f"Hello <@{user2['id']}>",
        ),
        dummy_message_factory(
            user_id=user2["id"],
            text="foobar",
            reactions=copy.deepcopy(reactions[:1]),
        ),
        dummy_message_factory(
            user_id=user3["id"],
            text=f"Is this @{user5['id']} a user mention?",
        ),
        dummy_message_factory(
            user_id=user4["id"],
            text="hello",
            reactions=copy.deepcopy(reactions),
        ),
    ]

    mock_user_retrieval_slack_client.users_list.side_effect = [
        {
            "ok": True,
            "members": [user1, user2, user3],
            "response_metadata": {"next_cursor": "cursor1"},
        },
        {
            "ok": True,
            "members": [user4, user5],
            "response_metadata": {"next_cursor": None},
        },
    ]

    response = await populate_users_in_messages(
        auth_token=mock_context.get_auth_token_or_empty(),
        messages=messages,
    )

    msg1 = response[0]
    msg2 = response[1]
    msg3 = response[2]
    msg4 = response[3]

    assert msg1["user"] == user1_short
    assert msg1["text"] == f"Hello {user2_mention}"
    assert "reactions" not in msg1

    assert msg2["user"] == user2_short
    assert msg2["text"] == "foobar"
    assert "reactions" in msg2
    assert len(msg2["reactions"]) == 1
    assert msg2["reactions"][0]["name"] == "thumbsup"
    assert msg2["reactions"][0]["users"] == [user1_short, user2_short]

    assert msg3["user"] == user3_short
    assert msg3["text"] == f"Is this @{user5['id']} a user mention?"
    assert "reactions" not in msg3
    assert user5_mention not in msg3["text"]

    assert msg4["user"] == user4_short
    assert msg4["text"] == "hello"
    assert "reactions" in msg4
    assert len(msg4["reactions"]) == 2
    assert msg4["reactions"][0]["name"] == "thumbsup"
    assert msg4["reactions"][0]["users"] == [user1_short, user2_short]
    assert msg4["reactions"][1]["name"] == "partyparrot"
    assert msg4["reactions"][1]["users"] == [user3_short, user4_short]



from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_slack
from arcade_slack.tools.chat import get_users_in_conversation

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.8,
    warn_threshold=0.9,
)


catalog = ToolCatalog()
# Register the Slack tools
catalog.add_module(arcade_slack)


@tool_eval()
def get_users_in_conversation_eval_suite() -> EvalSuite:
    """Create an evaluation suite for tools getting conversations members."""
    suite = EvalSuite(
        name="Slack Tools Evaluation",
        system_message="You are an AI assistant that can interact with Slack to send messages and get information from conversations, users, etc.",
        catalog=catalog,
        rubric=rubric,
    )

    user_messages = [
        "Get the members of the #general channel",
        "Get the users in the #general channel",
        "Get a list of people in the #general channel",
        "Get a list of people in the general channel",
        "Show me who's in the #general channel",
        "Who is in the general channel?",
    ]

    for user_message in user_messages:
        suite.add_case(
            name=f"Get users in channel by channel name: {user_message}",
            user_message=user_message,
            expected_tool_calls=[
                ExpectedToolCall(
                    func=get_users_in_conversation,
                    args={
                        "conversation_id": None,
                        "channel_name": "general",
                    },
                ),
            ],
            critics=[
                BinaryCritic(critic_field="conversation_id", weight=0.4),
                BinaryCritic(critic_field="channel_name", weight=0.6),
            ],
        )

    suite.add_case(
        name="Get users in conversation by conversation id",
        user_message="Get the users in the conversation with id '1234567890'",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_users_in_conversation,
                args={
                    "conversation_id": "1234567890",
                    "channel_name": None,
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=0.6),
            BinaryCritic(critic_field="channel_name", weight=0.4),
        ],
    )

    return suite



import json

import pytest
from arcade_tdk.errors import RetryableToolError, ToolExecutionError
from slack_sdk.errors import SlackApiError

from arcade_slack.user_retrieval import (
    get_single_user_by_id,
    get_users_by_id,
    get_users_by_id_username_or_email,
)
from arcade_slack.utils import (
    cast_user_dict,
    extract_basic_user_info,
    short_user_info,
)


@pytest.mark.asyncio
async def test_get_multiple_users_by_emails_success(
    mock_context, mock_user_retrieval_slack_client, dummy_user_factory
):
    user1 = dummy_user_factory()
    user2 = dummy_user_factory()

    emails = [user1["profile"]["email"], user2["profile"]["email"]]

    mock_user_retrieval_slack_client.users_lookupByEmail.side_effect = [
        {"ok": True, "user": user1},
        {"ok": True, "user": user2},
    ]

    response = await get_users_by_id_username_or_email(context=mock_context, emails=emails)

    assert response == [extract_basic_user_info(user1), extract_basic_user_info(user2)]


@pytest.mark.asyncio
async def test_get_multiple_users_by_usernames_or_emails_with_emails_not_found(
    mock_context,
    mock_user_retrieval_slack_client,
    mock_users_slack_client,
    dummy_user_factory,
):
    user1 = dummy_user_factory(email="user1@example.com")

    emails = [user1["profile"]["email"], "not_found@example.com"]

    async def lookup_by_email_side_effect(*, email):
        if email == "user1@example.com":
            return {"ok": True, "user": user1}
        raise SlackApiError(
            message="User not found",
            response={"error": "user_not_found"},
        )

    mock_user_retrieval_slack_client.users_lookupByEmail.side_effect = lookup_by_email_side_effect

    mock_users_slack_client.users_list.return_value = {
        "ok": True,
        "members": [user1],
    }

    with pytest.raises(RetryableToolError) as error:
        await get_users_by_id_username_or_email(context=mock_context, emails=emails)

    assert "not_found@example.com" in error.value.message
    assert json.dumps(short_user_info(user1)) in error.value.additional_prompt_content


@pytest.mark.asyncio
async def test_get_multiple_users_by_usernames_or_emails_with_usernames_success(
    mock_context, mock_user_retrieval_slack_client, dummy_user_factory
):
    user1 = dummy_user_factory()
    user2 = dummy_user_factory()

    usernames = [user1["name"], user2["name"]]

    mock_user_retrieval_slack_client.users_list.return_value = {
        "ok": True,
        "members": [user1, user2],
    }

    response = await get_users_by_id_username_or_email(context=mock_context, usernames=usernames)

    assert response == [extract_basic_user_info(user1), extract_basic_user_info(user2)]


@pytest.mark.asyncio
async def test_get_multiple_users_by_usernames_or_emails_with_usernames_not_found(
    mock_context, mock_user_retrieval_slack_client, dummy_user_factory
):
    user1 = dummy_user_factory()
    user2 = dummy_user_factory()
    user3 = dummy_user_factory()

    usernames = [user1["name"], "username_not_found"]

    mock_user_retrieval_slack_client.users_list.return_value = {
        "ok": True,
        "members": [user1, user2, user3],
    }

    with pytest.raises(RetryableToolError) as error:
        await get_users_by_id_username_or_email(context=mock_context, usernames=usernames)

    assert "username_not_found" in error.value.message
    assert json.dumps(short_user_info(user1)) in error.value.additional_prompt_content
    assert json.dumps(short_user_info(user2)) in error.value.additional_prompt_content
    assert json.dumps(short_user_info(user3)) in error.value.additional_prompt_content


@pytest.mark.asyncio
async def test_get_multiple_users_by_mixed_usernames_and_emails_success(
    mock_context, mock_user_retrieval_slack_client, dummy_user_factory
):
    user1 = dummy_user_factory()
    user2 = dummy_user_factory()
    user3 = dummy_user_factory()
    user4 = dummy_user_factory()

    mock_user_retrieval_slack_client.users_list.return_value = {
        "ok": True,
        "members": [user1, user2],
    }
    mock_user_retrieval_slack_client.users_lookupByEmail.side_effect = [
        {"ok": True, "user": user3},
        {"ok": True, "user": user4},
    ]

    response = await get_users_by_id_username_or_email(
        context=mock_context,
        usernames=[user1["name"], user2["name"]],
        emails=[user3["profile"]["email"], user4["profile"]["email"]],
    )

    assert response == [
        extract_basic_user_info(user1),
        extract_basic_user_info(user2),
        extract_basic_user_info(user3),
        extract_basic_user_info(user4),
    ]


@pytest.mark.asyncio
async def test_get_single_user_by_id_success(
    mock_context, mock_user_retrieval_slack_client, dummy_user_factory
):
    user = dummy_user_factory()

    mock_user_retrieval_slack_client.users_info.return_value = {"ok": True, "user": user}

    response = await get_single_user_by_id(
        auth_token=mock_context.get_auth_token_or_empty(),
        user_id=user["id"],
    )

    assert response == cast_user_dict(user)


@pytest.mark.asyncio
async def test_get_single_user_by_id_not_found(
    mock_context, mock_user_retrieval_slack_client, dummy_user_factory
):
    user = dummy_user_factory()

    mock_user_retrieval_slack_client.users_info.side_effect = SlackApiError(
        message="User not found",
        response={"error": "user_not_found"},
    )

    response = await get_single_user_by_id(
        auth_token=mock_context.get_auth_token_or_empty(),
        user_id=user["id"],
    )

    assert response is None


@pytest.mark.asyncio
async def test_get_single_user_by_id_not_ok(
    mock_context, mock_user_retrieval_slack_client, dummy_user_factory
):
    user = dummy_user_factory()

    mock_user_retrieval_slack_client.users_info.return_value = {"ok": False, "error": "not_ok"}

    response = await get_single_user_by_id(
        auth_token=mock_context.get_auth_token_or_empty(),
        user_id=user["id"],
    )

    assert response is None


@pytest.mark.asyncio
async def test_get_single_user_by_id_unknown_error(
    mock_context, mock_user_retrieval_slack_client, dummy_user_factory
):
    user = dummy_user_factory()

    mock_user_retrieval_slack_client.users_info.side_effect = SlackApiError(
        message="Unknown error",
        response={"error": "unknown_error_string"},
    )

    with pytest.raises(ToolExecutionError) as error:
        await get_single_user_by_id(
            auth_token=mock_context.get_auth_token_or_empty(),
            user_id=user["id"],
        )

    assert user["id"] in error.value.message
    assert "unknown_error_string" in error.value.developer_message


@pytest.mark.asyncio
async def test_get_users_by_id_with_one_user_id_success(
    mock_context, mock_user_retrieval_slack_client, dummy_user_factory
):
    user1 = dummy_user_factory()

    mock_user_retrieval_slack_client.users_info.return_value = {"ok": True, "user": user1}

    response = await get_users_by_id(
        auth_token=mock_context.get_auth_token_or_empty(),
        user_ids=[user1["id"]],
    )

    assert response == {"users": [cast_user_dict(user1)], "not_found": []}

    mock_user_retrieval_slack_client.users_list.assert_not_called()


@pytest.mark.asyncio
async def test_get_users_by_id_with_one_user_id_not_found(
    mock_context, mock_user_retrieval_slack_client, dummy_user_factory
):
    user1 = dummy_user_factory()

    mock_user_retrieval_slack_client.users_info.side_effect = SlackApiError(
        message="User not found",
        response={"error": "user_not_found"},
    )

    response = await get_users_by_id(
        auth_token=mock_context.get_auth_token_or_empty(),
        user_ids=[user1["id"]],
    )

    assert response == {"users": [], "not_found": [user1["id"]]}

    mock_user_retrieval_slack_client.users_list.assert_not_called()


@pytest.mark.asyncio
async def test_get_users_by_id_with_multiple_user_ids_success(
    mock_context, mock_user_retrieval_slack_client, dummy_user_factory
):
    user1 = dummy_user_factory()
    user2 = dummy_user_factory()
    user3 = dummy_user_factory()
    user4 = dummy_user_factory()

    mock_user_retrieval_slack_client.users_list.side_effect = [
        {
            "ok": True,
            "members": [user1, user2],
            "response_metadata": {"next_cursor": "next_cursor"},
        },
        {"ok": True, "members": [user3, user4]},
    ]

    response = await get_users_by_id(
        auth_token=mock_context.get_auth_token_or_empty(),
        user_ids=[user1["id"], user4["id"]],
    )

    assert response == {"users": [cast_user_dict(user1), cast_user_dict(user4)], "not_found": []}


@pytest.mark.asyncio
async def test_get_users_by_id_with_multiple_user_ids_some_not_found(
    mock_context, mock_user_retrieval_slack_client, dummy_user_factory
):
    user1 = dummy_user_factory(id_="U1")
    user2 = dummy_user_factory(id_="U2")
    user3 = dummy_user_factory(id_="U3")
    user4 = dummy_user_factory(id_="U4")

    mock_user_retrieval_slack_client.users_list.side_effect = [
        {
            "ok": True,
            "members": [user1, user2],
            "response_metadata": {"next_cursor": "next_cursor"},
        },
        {
            "ok": True,
            "members": [user3, user4],
            "response_metadata": {"next_cursor": None},
        },
    ]

    response = await get_users_by_id(
        auth_token=mock_context.get_auth_token_or_empty(),
        user_ids=[user1["id"], user4["id"], "user_not_exists"],
    )

    assert response == {
        "users": [cast_user_dict(user1), cast_user_dict(user4)],
        "not_found": ["user_not_exists"],
    }



import json
from unittest.mock import patch

import pytest
from arcade_tdk.errors import RetryableToolError, ToolExecutionError
from slack_sdk.errors import SlackApiError

from arcade_slack.tools.users import (
    get_users_info,
    list_users,
)
from arcade_slack.utils import extract_basic_user_info, short_user_info


@pytest.mark.asyncio
async def test_get_user_info_by_id_success(mock_context, mock_user_retrieval_slack_client):
    # Mock the response from slackClient.users_info
    mock_user = {
        "id": "U12345",
        "name": "testuser",
        "real_name": "Test User",
        "profile": {"email": "testuser@example.com"},
    }
    mock_user_retrieval_slack_client.users_info.return_value = {"ok": True, "user": mock_user}

    response = await get_users_info(mock_context, user_ids=["U12345"])

    mock_user_retrieval_slack_client.users_info.assert_called_once_with(user="U12345")

    expected_response = extract_basic_user_info(mock_user)
    assert response == {"users": [expected_response]}


@pytest.mark.asyncio
@patch("arcade_slack.tools.users.list_users")
async def test_get_user_info_by_id_user_not_found(
    mock_list_users, mock_context, mock_user_retrieval_slack_client
):
    error_response = {"ok": False, "error": "user_not_found"}
    mock_user_retrieval_slack_client.users_info.side_effect = SlackApiError(
        message="User not found",
        response=error_response,
    )

    existing_user = {"id": "U12345", "name": "testuser"}
    mock_list_users.return_value = {"users": [existing_user]}
    mock_list_users.__tool_name__ = list_users.__tool_name__

    with pytest.raises(RetryableToolError) as e:
        await get_users_info(mock_context, user_ids=["U99999"])

        assert existing_user["id"] in e.value.additional_prompt_content

    mock_user_retrieval_slack_client.users_info.assert_called_once_with(user="U99999")
    mock_list_users.assert_called_once_with(mock_context, limit=100, exclude_bots=True)


@pytest.mark.asyncio
async def test_list_users_success(mock_context, mock_users_slack_client):
    mock_users_slack_client.users_list.return_value = {
        "ok": True,
        "members": [{"id": "U12345"}],
    }
    response = await list_users(mock_context)
    assert response == {
        "users": [extract_basic_user_info({"id": "U12345"})],
        "next_cursor": None,
    }


@pytest.mark.asyncio
async def test_list_users_with_pagination_success(
    dummy_user_factory, mock_context, mock_users_slack_client
):
    user1 = dummy_user_factory()
    user2 = dummy_user_factory()
    user3 = dummy_user_factory()
    user4 = dummy_user_factory()

    mock_users_slack_client.users_list.side_effect = [
        {
            "ok": True,
            "members": [user1, user2],
            "response_metadata": {"next_cursor": "cursor_xyz"},
        },
        {
            "ok": True,
            "members": [user3, user4],
            "response_metadata": {"next_cursor": None},
        },
    ]
    response = await list_users(mock_context, limit=3)
    assert response == {
        "users": [
            extract_basic_user_info(user1),
            extract_basic_user_info(user2),
            extract_basic_user_info(user3),
            extract_basic_user_info(user4),
        ],
        "next_cursor": None,
    }

    assert mock_users_slack_client.users_list.call_count == 2


@pytest.mark.asyncio
async def test_get_user_by_username_success(
    mock_context,
    mock_user_retrieval_slack_client,
    dummy_user_factory,
):
    user1 = dummy_user_factory()
    user2 = dummy_user_factory()

    mock_user_retrieval_slack_client.users_list.return_value = {
        "ok": True,
        "members": [user1, user2],
    }

    response = await get_users_info(mock_context, usernames=[user1["name"]])

    assert response == {"users": [extract_basic_user_info(user1)]}


@pytest.mark.asyncio
async def test_get_user_by_username_with_pagination_success(
    mock_context, mock_user_retrieval_slack_client, dummy_user_factory
):
    user1 = dummy_user_factory()
    user2 = dummy_user_factory()
    user3 = dummy_user_factory()
    user4 = dummy_user_factory()
    user5 = dummy_user_factory()

    mock_user_retrieval_slack_client.users_list.side_effect = [
        {
            "ok": True,
            "members": [user1, user2],
            "response_metadata": {"next_cursor": "cursor1"},
        },
        {
            "ok": True,
            "members": [user3, user4],
            "response_metadata": {"next_cursor": "cursor2"},
        },
        {
            "ok": True,
            "members": [user5],
            "response_metadata": {"next_cursor": None},
        },
    ]

    response = await get_users_info(mock_context, usernames=[user3["name"]])

    assert response == {"users": [extract_basic_user_info(user3)]}

    assert mock_user_retrieval_slack_client.users_list.call_count == 2


@pytest.mark.asyncio
async def test_get_user_by_username_not_found(
    mock_context, mock_user_retrieval_slack_client, dummy_user_factory
):
    user1 = dummy_user_factory()
    user2 = dummy_user_factory()
    user3 = dummy_user_factory(is_bot=True)

    mock_user_retrieval_slack_client.users_list.return_value = {
        "ok": True,
        "members": [user1, user2, user3],
    }

    with pytest.raises(RetryableToolError) as e:
        await get_users_info(mock_context, usernames=[user1["name"] + "not_found"])

    # Check that the error message contains the available users
    assert user1["id"] in e.value.additional_prompt_content
    assert user2["id"] in e.value.additional_prompt_content
    assert user3["id"] not in e.value.additional_prompt_content


@pytest.mark.asyncio
async def test_get_multiple_users_by_username_success(
    mock_context, mock_user_retrieval_slack_client, dummy_user_factory
):
    user1 = dummy_user_factory()
    user2 = dummy_user_factory()
    user3 = dummy_user_factory()

    mock_user_retrieval_slack_client.users_list.return_value = {
        "ok": True,
        "members": [user1, user2, user3],
    }

    response = await get_users_info(
        mock_context,
        usernames=[user1["name"], user2["name"]],
    )

    assert response == {
        "users": [
            extract_basic_user_info(user1),
            extract_basic_user_info(user2),
        ]
    }


@pytest.mark.asyncio
async def test_get_multiple_users_by_username_with_pagination_success(
    mock_context, mock_user_retrieval_slack_client, dummy_user_factory
):
    user1 = dummy_user_factory()
    user2 = dummy_user_factory()
    user3 = dummy_user_factory()
    user4 = dummy_user_factory()
    user5 = dummy_user_factory()

    mock_user_retrieval_slack_client.users_list.side_effect = [
        {"ok": True, "members": [user1, user2], "response_metadata": {"next_cursor": "cursor1"}},
        {"ok": True, "members": [user3, user4], "response_metadata": {"next_cursor": "cursor2"}},
        {"ok": True, "members": [user5], "response_metadata": {"next_cursor": None}},
    ]

    response = await get_users_info(mock_context, usernames=[user1["name"], user3["name"]])

    assert response == {"users": [extract_basic_user_info(user1), extract_basic_user_info(user3)]}
    assert mock_user_retrieval_slack_client.users_list.call_count == 2


@pytest.mark.asyncio
async def test_get_multiple_users_by_username_not_found(
    mock_context, mock_user_retrieval_slack_client, dummy_user_factory
):
    user1 = dummy_user_factory()
    user2 = dummy_user_factory()
    user3 = dummy_user_factory(is_bot=True)

    mock_user_retrieval_slack_client.users_list.return_value = {
        "ok": True,
        "members": [user1, user2, user3],
    }

    not_found_username = f"{user2['name']} not_found"

    with pytest.raises(RetryableToolError) as e:
        await get_users_info(mock_context, usernames=[user1["name"], not_found_username])

    assert user1["id"] in e.value.additional_prompt_content
    assert user2["id"] in e.value.additional_prompt_content
    assert user3["id"] not in e.value.additional_prompt_content


@pytest.mark.asyncio
async def test_get_user_by_email_success(
    mock_context, mock_user_retrieval_slack_client, dummy_user_factory
):
    user = dummy_user_factory()
    mock_user_retrieval_slack_client.users_lookupByEmail.return_value = {
        "ok": True,
        "user": user,
    }
    response = await get_users_info(mock_context, emails=[user["profile"]["email"]])
    assert response == {"users": [extract_basic_user_info(user)]}


@pytest.mark.asyncio
async def test_get_user_by_email_not_found(
    mock_context, mock_users_slack_client, mock_user_retrieval_slack_client, dummy_user_factory
):
    additional_user = dummy_user_factory(email="additional_user@example.com")

    async def lookup_by_email_side_effect(*, email):
        if email == "additional_user@example.com":
            return {"ok": True, "user": additional_user}
        raise SlackApiError(
            message="User not found",
            response={"ok": False, "error": "user_not_found"},
        )

    mock_user_retrieval_slack_client.users_lookupByEmail.side_effect = lookup_by_email_side_effect
    mock_users_slack_client.users_list.return_value = {
        "ok": True,
        "members": [additional_user],
    }

    with pytest.raises(RetryableToolError) as e:
        await get_users_info(mock_context, emails=["not_found@example.com"])

    assert "not_found@example.com" in e.value.message
    assert json.dumps(short_user_info(additional_user)) in e.value.additional_prompt_content


@pytest.mark.asyncio
async def test_get_multiple_users_by_email_success(
    mock_context, mock_user_retrieval_slack_client, dummy_user_factory
):
    user1 = dummy_user_factory()
    user2 = dummy_user_factory()
    user3 = dummy_user_factory()

    mock_user_retrieval_slack_client.users_lookupByEmail.side_effect = [
        {"ok": True, "user": user1},
        {"ok": True, "user": user2},
        {"ok": True, "user": user3},
    ]

    response = await get_users_info(
        mock_context,
        emails=[
            user1["profile"]["email"],
            user2["profile"]["email"],
            user3["profile"]["email"],
        ],
    )

    assert response == {
        "users": [
            extract_basic_user_info(user1),
            extract_basic_user_info(user2),
            extract_basic_user_info(user3),
        ]
    }


@pytest.mark.asyncio
async def test_get_multiple_users_by_email_with_invalid_address(
    mock_context, mock_user_retrieval_slack_client
):
    with pytest.raises(ToolExecutionError) as e:
        await get_users_info(
            mock_context,
            emails=["amyra@jades.com", "invalid_address"],
        )

    mock_user_retrieval_slack_client.assert_not_called()
    assert e.value.message == "Invalid email address: invalid_address"


@pytest.mark.asyncio
async def test_get_multiple_users_by_email_not_found(
    mock_context, mock_user_retrieval_slack_client, mock_users_slack_client, dummy_user_factory
):
    user1 = dummy_user_factory()
    user2 = dummy_user_factory()

    async def lookup_by_email_side_effect(*, email):
        if email == user1["profile"]["email"]:
            return {"ok": True, "user": user1}
        if email == user2["profile"]["email"]:
            return {"ok": True, "user": user2}
        raise SlackApiError(
            message="User not found",
            response={"ok": False, "error": "user_not_found"},
        )

    mock_user_retrieval_slack_client.users_lookupByEmail.side_effect = lookup_by_email_side_effect
    mock_users_slack_client.users_list.return_value = {
        "ok": True,
        "members": [user1, user2],
    }

    with pytest.raises(RetryableToolError) as e:
        await get_users_info(
            mock_context,
            emails=[
                user1["profile"]["email"],
                user2["profile"]["email"],
                "not_found@example.com",
            ],
        )

    assert "not_found@example.com" in e.value.message
    assert json.dumps(short_user_info(user1)) in e.value.additional_prompt_content
    assert json.dumps(short_user_info(user2)) in e.value.additional_prompt_content



import json
from typing import cast

from arcade_tdk.errors import RetryableToolError, ToolExecutionError
from slack_sdk.errors import SlackApiError
from slack_sdk.web.async_client import AsyncWebClient

from arcade_slack.models import (
    ConversationType,
    FindChannelByNameSentinel,
)
from arcade_slack.utils import (
    async_paginate,
    extract_conversation_metadata,
)


async def get_conversation_by_id(
    auth_token: str,
    conversation_id: str,
) -> dict:
    """Get metadata of a conversation in Slack by the conversation_id."""
    try:
        slack_client = AsyncWebClient(token=auth_token)
        response = await slack_client.conversations_info(
            channel=conversation_id,
            include_locale=True,
            include_num_members=True,
        )
        return dict(**extract_conversation_metadata(response["channel"]))

    except SlackApiError as e:
        slack_error = cast(str, e.response.get("error", ""))
        if "not_found" in slack_error.lower():
            message = f"Conversation with ID '{conversation_id}' not found."
            raise ToolExecutionError(message=message, developer_message=message)
        raise


async def get_channel_by_name(
    auth_token: str,
    channel_name: str,
) -> dict:
    channel_name_casefolded = channel_name.lstrip("#").casefold()

    slack_client = AsyncWebClient(token=auth_token)

    results, _ = await async_paginate(
        func=slack_client.conversations_list,
        response_key="channels",
        types=",".join([
            ConversationType.PUBLIC_CHANNEL.value,
            ConversationType.PRIVATE_CHANNEL.value,
        ]),
        exclude_archived=True,
        sentinel=FindChannelByNameSentinel(channel_name_casefolded),
    )

    available_channels = []

    for channel in results:
        if channel["name"].casefold() == channel_name_casefolded:
            return dict(**extract_conversation_metadata(channel))
        else:
            available_channels.append({"id": channel["id"], "name": channel["name"]})

    error_message = f"Channel with name '{channel_name}' not found."

    raise RetryableToolError(
        message=error_message,
        developer_message=error_message,
        additional_prompt_content=f"Available channels: {json.dumps(available_channels)}",
        retry_after_ms=500,
    )



import json

from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    SimilarityCritic,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_slack
from arcade_slack.tools.chat import send_message

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.8,
    warn_threshold=0.9,
)


catalog = ToolCatalog()
# Register the Slack tools
catalog.add_module(arcade_slack)


@tool_eval()
def send_message_eval_suite() -> EvalSuite:
    """Create an evaluation suite for Slack messaging tools."""
    suite = EvalSuite(
        name="Slack Messaging Tools Evaluation",
        system_message="You are an AI assistant that can send direct messages and post messages to channels in Slack using the provided tools.",
        catalog=catalog,
        rubric=rubric,
    )

    # Send DM to User Scenarios
    suite.add_case(
        name="Send DM to user with clear username",
        user_message="Send a direct message to johndoe saying 'Hello, can we meet at 3 PM?'",
        expected_tool_calls=[
            ExpectedToolCall(
                func=send_message,
                args={
                    "message": "Hello, can we meet at 3 PM?",
                    "conversation_id": None,
                    "channel_name": None,
                    "user_ids": None,
                    "emails": None,
                    "usernames": ["johndoe"],
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="message", weight=0.3, similarity_threshold=0.9),
            BinaryCritic(critic_field="conversation_id", weight=0.1),
            BinaryCritic(critic_field="channel_name", weight=0.1),
            BinaryCritic(critic_field="user_ids", weight=0.1),
            BinaryCritic(critic_field="emails", weight=0.1),
            BinaryCritic(critic_field="usernames", weight=0.3),
        ],
    )

    suite.add_case(
        name="Send message to channel with clear name",
        user_message="Post 'The new feature is now live!' in the #announcements channel",
        expected_tool_calls=[
            ExpectedToolCall(
                func=send_message,
                args={
                    "message": "The new feature is now live!",
                    "conversation_id": None,
                    "channel_name": "announcements",
                    "user_ids": None,
                    "emails": None,
                    "usernames": None,
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="message", weight=0.3, similarity_threshold=0.9),
            BinaryCritic(critic_field="conversation_id", weight=0.1),
            BinaryCritic(critic_field="channel_name", weight=0.3),
            BinaryCritic(critic_field="user_ids", weight=0.1),
            BinaryCritic(critic_field="emails", weight=0.1),
            BinaryCritic(critic_field="usernames", weight=0.1),
        ],
    )

    suite.add_case(
        name="Send message to channel with ambiguous name",
        user_message="Inform the team in the general channel about the upcoming maintenance",
        expected_tool_calls=[
            ExpectedToolCall(
                func=send_message,
                args={
                    "message": "Attention team: There will be upcoming maintenance. Please save your work and expect some downtime.",
                    "conversation_id": None,
                    "channel_name": "general",
                    "user_ids": None,
                    "emails": None,
                    "usernames": None,
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="message", weight=0.3, similarity_threshold=0.9),
            BinaryCritic(critic_field="conversation_id", weight=0.1),
            BinaryCritic(critic_field="channel_name", weight=0.3),
            BinaryCritic(critic_field="user_ids", weight=0.1),
            BinaryCritic(critic_field="emails", weight=0.1),
            BinaryCritic(critic_field="usernames", weight=0.1),
        ],
    )

    # Adversarial Scenarios
    suite.add_case(
        name="Ambiguous between DM and channel message",
        user_message="general",
        expected_tool_calls=[
            ExpectedToolCall(
                func=send_message,
                args={
                    "message": "Great job on the presentation!",
                    "conversation_id": None,
                    "channel_name": "general",
                    "user_ids": None,
                    "emails": None,
                    "usernames": None,
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="message", weight=0.3, similarity_threshold=0.9),
            BinaryCritic(critic_field="conversation_id", weight=0.1),
            BinaryCritic(critic_field="channel_name", weight=0.3),
            BinaryCritic(critic_field="user_ids", weight=0.1),
            BinaryCritic(critic_field="emails", weight=0.1),
            BinaryCritic(critic_field="usernames", weight=0.1),
        ],
        additional_messages=[
            {"role": "user", "content": "Send 'Great job on the presentation!' to the team"},
            {
                "role": "assistant",
                "content": "",
                "tool_calls": [
                    {
                        "id": "call_1",
                        "type": "function",
                        "function": {
                            "name": "Slack_ListConversationsMetadata",
                            "arguments": '{"limit":20, "conversation_types":["public_channel"]}',
                        },
                    }
                ],
            },
            {
                "role": "tool",
                "content": json.dumps({
                    "conversations": [
                        {
                            "conversation_type": "public_channel",
                            "id": "channel1",
                            "is_archived": False,
                            "is_member": True,
                            "is_private": False,
                            "name": "random",
                            "num_members": 999,
                            "purpose": "Random stuff",
                        },
                        {
                            "conversation_type": "public_channel",
                            "id": "channel2",
                            "is_archived": False,
                            "is_member": True,
                            "is_private": False,
                            "name": "general",
                            "num_members": 999,
                            "purpose": "Just a general channel",
                        },
                    ],
                    "next_cursor": "",
                }),
                "tool_call_id": "call_1",
                "name": "Slack_ListPublicChannelsMetadata",
            },
            {
                "role": "assistant",
                "content": 'To send the message "Great job on the presentation!" to the team, please let me know which Slack channel you\'d like to use:\n\n1. #random\n2. #general\n\nPlease let me know your choice!',
            },
        ],
    )

    suite.add_case(
        name="Multiple recipients in DM request",
        user_message="Send DMs to the users 'alice' and 'bob' about rescheduling our meeting tomorrow. I have too much work to do.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=send_message,
                args={
                    "conversation_id": None,
                    "channel_name": None,
                    "user_ids": None,
                    "emails": None,
                    "usernames": ["alice"],
                },
            ),
            ExpectedToolCall(
                func=send_message,
                args={
                    "conversation_id": None,
                    "channel_name": None,
                    "user_ids": None,
                    "emails": None,
                    "usernames": ["bob"],
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=0.1),
            BinaryCritic(critic_field="channel_name", weight=0.1),
            BinaryCritic(critic_field="user_ids", weight=0.1),
            BinaryCritic(critic_field="emails", weight=0.1),
            BinaryCritic(critic_field="usernames", weight=0.6),
        ],
    )

    suite.add_case(
        name="Multiple recipients in MPIM request",
        user_message="Send a message to the users 'alice' and 'bob' about rescheduling our meeting tomorrow. I have too much work to do.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=send_message,
                args={
                    "conversation_id": None,
                    "channel_name": None,
                    "user_ids": None,
                    "emails": None,
                    "usernames": ["alice", "bob"],
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="conversation_id", weight=0.1),
            BinaryCritic(critic_field="channel_name", weight=0.1),
            BinaryCritic(critic_field="user_ids", weight=0.1),
            BinaryCritic(critic_field="emails", weight=0.1),
            BinaryCritic(critic_field="usernames", weight=0.6),
        ],
    )

    suite.add_case(
        name="Channel name similar to username",
        user_message="Post 'sounds great!' in john-project channel",
        expected_tool_calls=[
            ExpectedToolCall(
                func=send_message,
                args={
                    "conversation_id": None,
                    "channel_name": "john-project",
                    "user_ids": None,
                    "emails": None,
                    "usernames": None,
                    "message": "Sounds great!",
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="message", weight=0.3, similarity_threshold=0.9),
            BinaryCritic(critic_field="conversation_id", weight=0.1),
            BinaryCritic(critic_field="channel_name", weight=0.3),
            BinaryCritic(critic_field="user_ids", weight=0.1),
            BinaryCritic(critic_field="emails", weight=0.1),
            BinaryCritic(critic_field="usernames", weight=0.1),
        ],
    )

    return suite



import asyncio
from unittest.mock import AsyncMock

import pytest
from arcade_tdk.errors import ToolExecutionError
from slack_sdk.errors import SlackApiError

from arcade_slack.models import (
    ConcurrencySafeCoroutineCaller,
    FindChannelByNameSentinel,
    FindMultipleUsersByUsernameSentinel,
    FindUserByUsernameSentinel,
    GetUserByEmailCaller,
)


def test_find_user_by_username_sentinel():
    sentinel = FindUserByUsernameSentinel(username="jenifer")
    assert sentinel(last_result=[{"name": "jack"}]) is False
    assert sentinel(last_result=[{"name": "john"}, {"name": "jack"}]) is False
    assert sentinel(last_result=[{"name": "hello"}, {"name": "jenifer"}]) is True
    assert sentinel(last_result=[{"name": "JENIFER"}]) is True


def test_find_multiple_users_by_username_sentinel():
    sentinel = FindMultipleUsersByUsernameSentinel(usernames=["jenifer", "jack"])
    assert sentinel(last_result=[{"name": "jack"}]) is False
    assert sentinel(last_result=[{"name": "john"}, {"name": "jack"}]) is False
    assert sentinel(last_result=[{"name": "hello"}, {"name": "JENIFER"}]) is True
    assert sentinel(last_result=[{"name": "world"}]) is True


def test_find_channel_by_name_sentinel():
    sentinel = FindChannelByNameSentinel(channel_name="foobar")
    assert sentinel(last_result=[{"name": "foo"}]) is False
    assert sentinel(last_result=[{"name": "foo"}, {"name": "bar"}]) is False
    assert sentinel(last_result=[{"name": "foo"}, {"name": "foobar"}]) is True
    assert sentinel(last_result=[{"name": "FOObar"}]) is True


@pytest.mark.asyncio
async def test_concurrency_safe_coroutine_caller():
    mock_func = AsyncMock()
    mock_semaphore = AsyncMock(spec=asyncio.Semaphore)

    caller = ConcurrencySafeCoroutineCaller(mock_func, "arg1", "arg2", kwarg1="kwarg1")
    response = await caller(semaphore=mock_semaphore)

    assert response == mock_func.return_value
    mock_func.assert_awaited_once_with("arg1", "arg2", kwarg1="kwarg1")
    mock_semaphore.__aenter__.assert_awaited_once()
    mock_semaphore.__aexit__.assert_awaited_once()


@pytest.mark.asyncio
async def test_get_user_by_email_caller_success():
    mock_func = AsyncMock()
    mock_func.return_value = {"user": {"id": "U1234567890", "name": "John Doe"}}
    mock_semaphore = AsyncMock(spec=asyncio.Semaphore)

    caller = GetUserByEmailCaller(mock_func, "test@example.com")
    response = await caller(semaphore=mock_semaphore)

    assert response == {
        "user": {"id": "U1234567890", "name": "John Doe"},
        "email": "test@example.com",
    }
    mock_func.assert_awaited_once_with(email="test@example.com")
    mock_semaphore.__aenter__.assert_awaited_once()
    mock_semaphore.__aexit__.assert_awaited_once()


@pytest.mark.asyncio
async def test_get_user_by_email_caller_not_found_error():
    mock_func = AsyncMock(
        side_effect=SlackApiError(message="User not found", response={"error": "user_not_found"})
    )
    mock_semaphore = AsyncMock(spec=asyncio.Semaphore)

    caller = GetUserByEmailCaller(mock_func, "test@example.com")
    response = await caller(semaphore=mock_semaphore)

    assert response == {
        "user": None,
        "email": "test@example.com",
    }
    mock_func.assert_awaited_once_with(email="test@example.com")
    mock_semaphore.__aenter__.assert_awaited_once()
    mock_semaphore.__aexit__.assert_awaited_once()


@pytest.mark.asyncio
async def test_get_user_by_email_caller_unknown_slack_api_error():
    mock_func = AsyncMock(
        side_effect=SlackApiError(message="Unknown error", response={"error": "unknown_error"})
    )
    mock_semaphore = AsyncMock(spec=asyncio.Semaphore)

    caller = GetUserByEmailCaller(mock_func, "test@example.com")
    with pytest.raises(ToolExecutionError) as error:
        await caller(semaphore=mock_semaphore)

    assert error.value.message == "Error getting user by email"
    assert error.value.developer_message == "Error getting user by email: unknown_error"
    mock_func.assert_awaited_once_with(email="test@example.com")
    mock_semaphore.__aenter__.assert_awaited_once()
    mock_semaphore.__aexit__.assert_awaited_once()
